@rem SetupMgrTag
@echo off

rem
rem This is a SAMPLE batch script generated by Setup Manager.
rem If this script is moved from the location where it was generated, it may have to be modified.
rem

set AnswerFile=.\winnt.sif
set SetupFiles=D:\i386

D:\i386\winnt32 /s:%SetupFiles% /unattend:%AnswerFile%
